"""
This file serves a a 'proxy' for various order models,
but the source for these models has now been moved into separate files
"""

# Pass PurchaseOrder models through
from inventree.purchase_order import PurchaseOrder  # noqa:F401
from inventree.purchase_order import PurchaseOrderExtraLineItem  # noqa:F401
from inventree.purchase_order import PurchaseOrderLineItem  # noqa:F401
# Pass ReturnOrder models through
from inventree.return_order import ReturnOrder  # noqa:F401
from inventree.return_order import ReturnOrderExtraLineItem  # noqa:F401
from inventree.return_order import ReturnOrderLineItem  # noqa:F401
# Pass SalesOrder models through
from inventree.sales_order import SalesOrder  # noqa:F401
from inventree.sales_order import SalesOrderExtraLineItem  # noqa:F401
from inventree.sales_order import SalesOrderLineItem  # noqa:F401
from inventree.sales_order import SalesOrderShipment  # noqa:F401



# -*- coding: utf-8 -*-

import os
import sys

try:
    import Image
except ImportError:
    from PIL import Image

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.base import Attachment  # noqa: E402
from inventree import company  # noqa: E402
from inventree.part import Part  # noqa: E402


class ContactTest(InvenTreeTestCase):
    """Tests for the 'Contact' model"""

    def test_contact_create(self):
        """Test that we can create a new contact"""

        cmp = company.Company.list(self.api, limit=1)[0]
        n = company.Contact.count(self.api)

        for idx in range(3):
            company.Contact.create(self.api, data={
                'company': cmp.pk,
                'name': f"Random Name {idx}",
                'role': f"Random Role {idx}",
            })

        self.assertEqual(company.Contact.count(self.api), n + 3)


class CompanyTest(InvenTreeTestCase):
    """
    Test that Company related objects can be managed via the API
    """

    def test_fields(self):
        """
        Test field names via OPTIONS request
        """

        field_names = company.Company.fieldNames(self.api)

        for field in [
            'name',
            'image',
            'is_customer',
            'is_manufacturer',
            'is_supplier'
        ]:
            self.assertIn(field, field_names)

    def test_company_create(self):
        c = company.Company.create(self.api, {
            'name': 'Company',
        })

        self.assertIsNotNone(c)

    def test_company_parts(self):
        """
        Tests that the 'supplied' and 'manufactured' parts can be retrieved
        """

        c = company.Company.create(self.api, {
            'name': 'MyTestCompany',
            'description': 'A manufacturer *AND* a supplier',
            'is_manufacturer': True,
            'is_supplier': True,
        })

        self.assertIsNotNone(c)

        self.assertEqual(len(c.getManufacturedParts()), 0)
        self.assertEqual(len(c.getSuppliedParts()), 0)

        # Create some 'manufactured parts'
        for i in range(3):

            mpn = f"MPN_XYX-{i}_{c.pk}"
            sku = f"SKU_ABC-{i}_{c.pk}"

            # Create a new ManufacturerPart
            m_part = company.ManufacturerPart.create(self.api, {
                'part': 1,
                'manufacturer': c.pk,
                'MPN': mpn
            })

            # Creating a unique SupplierPart should also create a ManufacturerPart
            company.SupplierPart.create(self.api, {
                'supplier': c.pk,
                'part': 1,
                'manufacturer_part': m_part.pk,
                'SKU': sku,
            })

        self.assertEqual(len(c.getManufacturedParts()), 3)
        self.assertEqual(len(c.getSuppliedParts()), 3)

    def test_manufacturer_part_create(self):

        manufacturer = company.Company(self.api, 7)

        n = len(manufacturer.getManufacturedParts())

        # Create a new manufacturer part with a unique name
        manufacturer_part = company.ManufacturerPart.create(self.api, {
            'manufacturer': manufacturer.pk,
            'MPN': f'MPN_TEST_{n}',
            'part': 3,
        })

        self.assertIsNotNone(manufacturer_part)
        self.assertEqual(manufacturer_part.manufacturer, manufacturer.pk)

        # Check that listing the manufacturer parts against this manufacturer has increased by 1
        man_parts = company.ManufacturerPart.list(self.api, manufacturer=manufacturer.pk)
        self.assertEqual(len(man_parts), n + 1)

    def test_supplier_part_create(self):
        """
        Test that we can create SupplierPart objects via the API
        """

        supplier = company.Company(self.api, 1)

        # Find a purchaseable part
        parts = Part.list(self.api, purchasable=True)

        if len(parts) > 0:
            prt = parts[0]
        else:
            prt = Part.create(self.api, {
                'name': 'My purchaseable part',
                'description': 'A purchasenable part we can use to make a SupplierPart',
                'category': 1,
                'purchaseable': True
            })

        n = len(company.SupplierPart.list(self.api))

        supplier_part = company.SupplierPart.create(self.api, {
            'supplier': supplier.pk,
            'SKU': f'SKU_TEST_{n}',
            'part': prt.pk,
        })

        self.assertIsNotNone(supplier_part)
        self.assertTrue(supplier_part.part, prt.pk)

        self.assertEqual(len(company.SupplierPart.list(self.api)), n + 1)

    def test_upload_company_image(self):
        """
        Check we can upload image against a company
        """

        # Grab the first company
        c = company.Company.list(self.api)[0]

        # Ensure the company does *not* have an image
        c.save(data={'image': None})
        c.reload()

        self.assertIsNone(c.image)

        # Test that trying to *download* a null image results in failure!
        with self.assertRaises(ValueError):
            c.downloadImage("downloaded.png")

        # Now, let's actually upload a real image
        # Upload as an im-memory imgae file
        img = Image.new('RGB', (128, 128), color='blue')
        img.save('dummy_image.png')

        response = c.uploadImage('dummy_image.png')

        self.assertTrue(response)

        with self.assertRaises(FileNotFoundError):
            c.uploadImage('ddddummy_image.png')

        with self.assertRaises(TypeError):
            c.uploadImage(1)

        with self.assertRaises(TypeError):
            c.uploadImage(None)

    def test_attachments(self):
        """Unit tests for attachments."""

        if self.api.api_version < Attachment.MIN_API_VERSION:
            return

        # Create a new manufacturer part, if one does not already exist
        mps = company.ManufacturerPart.list(self.api)

        if len(mps) > 0:
            mp = mps[0]
        else:
            mp = company.ManufacturerPart.create(self.api, {
                'manufacturer': 7,
                'part': 3,
                'MPN': 'M7P3',
            })

        # Initially, ensure there are no attachments associated with this part
        for attachment in mp.getAttachments():
            attachment.delete()

        attachments = mp.getAttachments()

        self.assertTrue(len(attachments) == 0)

        # Let's create one!

        # Grab the first available manufacturer part

        response = mp.uploadAttachment(
            os.path.join(os.path.dirname(__file__), 'attachment.txt'),
            comment='Uploading a new manufacturer part attachment',
        )

        self.assertIsNotNone(response)

        attachments = mp.getAttachments()

        self.assertEqual(len(attachments), 1)

        attachment = attachments[0]

        self.assertEqual(attachment['comment'], 'Uploading a new manufacturer part attachment')

        attachments = mp.getAttachments()

        self.assertEqual(len(attachments), 1)



"""
PurchaseOrder models
"""

import inventree.base
import inventree.company
import inventree.part
import inventree.report


class PurchaseOrder(
    inventree.base.AttachmentMixin,
    inventree.base.ParameterMixin,
    inventree.base.MetadataMixin,
    inventree.base.StatusMixin,
    inventree.report.ReportPrintingMixin,
    inventree.base.InventreeObject,
):
    """ Class representing the PurchaseOrder database model """

    URL = 'order/po/'
    MODEL_TYPE = 'purchaseorder'

    def getSupplier(self):
        """Return the supplier (Company) associated with this order"""
        return inventree.company.Company(self._api, self.supplier)

    def getContact(self):
        """Return the contact associated with this order"""
        if self.contact is not None:
            return inventree.company.Contact(self._api, self.contact)
        else:
            return None

    def getLineItems(self, **kwargs):
        """ Return the line items associated with this order """
        return PurchaseOrderLineItem.list(self._api, order=self.pk, **kwargs)

    def getExtraLineItems(self, **kwargs):
        """ Return the line items associated with this order """
        return PurchaseOrderExtraLineItem.list(self._api, order=self.pk, **kwargs)

    def addLineItem(self, **kwargs):
        """
        Create (and return) new PurchaseOrderLineItem object against this PurchaseOrder
        """

        kwargs['order'] = self.pk

        return PurchaseOrderLineItem.create(self._api, data=kwargs)

    def addExtraLineItem(self, **kwargs):
        """
        Create (and return) new PurchaseOrderExtraLineItem object against this PurchaseOrder
        """

        kwargs['order'] = self.pk

        return PurchaseOrderExtraLineItem.create(self._api, data=kwargs)

    def issue(self, **kwargs):
        """
        Issue the purchase order
        """

        # Return
        return self._statusupdate(status='issue', **kwargs)

    def hold(self, **kwargs):
        """
        Hold the purchase order
        """

        # Return
        return self._statusupdate(status='hold', **kwargs)

    def receiveAll(self, location, status=10):
        """
        Receive all of the purchase order items, into the given location.

        Note that the location may be overwritten if a destination is saved in the PO for the line item.

        By default, the status is set to OK (Code 10).

        To modify the defaults, use the arguments:
            status: Status code
                    10 OK
                    50 ATTENTION
                    55 DAMAGED
                    60 DESTROYED
                    65 REJECTED
                    70 LOST
                    75 QUARANTINED
                    85 RETURNED
        """

        # Check if location is a model - or try to get an integer
        try:
            location_id = location.pk
        except:  # noqa:E722
            location_id = int(location)

        # Prepare request data
        items = list()
        for li in self.getLineItems():
            quantity_to_receive = li.quantity - li.received
            # Make sure quantity > 0
            if quantity_to_receive > 0:
                items.append(
                    {
                        'line_item': li.pk,
                        'supplier_part': li.part,
                        'quantity': quantity_to_receive,
                        'status': status,
                        'location': location_id,
                    }
                )

        # If nothing is left, quit here
        if len(items) < 1:
            return None

        data = {
            'items': items,
            'location': location_id
        }

        # Set the url
        URL = f"{self.URL}/{self.pk}/receive/"

        # Send data
        response = self._api.post(URL, data)

        # Reload
        self.reload()

        # Return
        return response


class PurchaseOrderLineItem(
    inventree.base.InventreeObject,
    inventree.base.MetadataMixin,
):
    """ Class representing the PurchaseOrderLineItem database model """

    URL = 'order/po-line/'

    def getSupplierPart(self):
        """
        Return the SupplierPart associated with this PurchaseOrderLineItem
        """
        return inventree.company.SupplierPart(self._api, self.part)

    def getPart(self):
        """
        Return the Part referenced by the associated SupplierPart
        """
        return inventree.part.Part(self._api, self.getSupplierPart().part)

    def getOrder(self):
        """
        Return the PurchaseOrder to which this PurchaseOrderLineItem belongs
        """
        return PurchaseOrder(self._api, self.order)

    def receive(self, quantity=None, status=10, location=None, expiry_date=None, batch_code=None, serial_numbers=None):
        """
        Mark this line item as received.

        By default, receives all remaining items in the order, and puts them in the destination defined in the PO.
        The status is set to OK (Code 10).

        To modify the defaults, use the arguments:
            quantity: Number of units to receive. If None, will calculate the quantity not yet received and receive these.
            status: Status code
                    10 OK
                    50 ATTENTION
                    55 DAMAGED
                    60 DESTROYED
                    65 REJECTED
                    70 LOST
                    75 QUARANTINED
                    85 RETURNED
            location: Location ID, or a StockLocation item

        If given, the following arguments are also sent as parameters:
            expiry_date
            batch_code
            serial_numbers
        """

        if quantity is None:
            # Subtract number of already received lines from the order quantity
            quantity = self.quantity - self.received

        if location is None:
            location_id = self.destination
        else:
            # Check if location is a model - or try to get an integer
            try:
                location_id = location.pk
            except:  # noqa:E722
                location_id = int(location)

        item_data = {
            'line_item': self.pk,
            'supplier_part': self.part,
            'quantity': quantity,
            'status': status,
            'location': location_id
        }

        # Optional fields which may be set
        if expiry_date:
            item_data['expiry_date'] = expiry_date

        if batch_code:
            item_data['batch_code'] = batch_code
        
        if serial_numbers:
            item_data['serial_numbers'] = serial_numbers

        # Prepare request data
        data = {
            'items': [
                item_data,
            ],
            'location': location_id
        }

        # Set the url
        URL = f"{self.getOrder().URL}/{self.getOrder().pk}/receive/"

        # Send data
        response = self._api.post(URL, data)

        # Reload
        self.reload()

        # Return
        return response


class PurchaseOrderExtraLineItem(
    inventree.base.InventreeObject,
    inventree.base.MetadataMixin,
):
    """ Class representing the PurchaseOrderExtraLineItem database model """

    URL = 'order/po-extra-line/'

    def getOrder(self):
        """
        Return the PurchaseOrder to which this PurchaseOrderLineItem belongs
        """
        return PurchaseOrder(self._api, self.order)



"""
ReturnOrder models
"""

import inventree.base
import inventree.company
import inventree.part
import inventree.report
import inventree.stock


class ReturnOrder(
    inventree.base.AttachmentMixin,
    inventree.base.ParameterMixin,
    inventree.base.MetadataMixin,
    inventree.base.StatusMixin,
    inventree.report.ReportPrintingMixin,
    inventree.base.InventreeObject,
):
    """Class representing the ReturnOrder database model"""

    URL = 'order/ro/'
    MIN_API_VERSION = 104
    MODEL_TYPE = 'returnorder'

    def getCustomer(self):
        """Return the customer associated with this order"""
        return inventree.company.Company(self._api, self.customer)

    def getContact(self):
        """Return the contact associated with this order"""
        if self.contact is not None:
            return inventree.company.Contact(self._api, self.contact)
        else:
            return None

    def getLineItems(self, **kwargs):
        """Return line items associated with this order"""
        return ReturnOrderLineItem.list(self._api, order=self.pk, **kwargs)

    def addLineItem(self, **kwargs):
        """Create (and return) a new ReturnOrderLineItem against this order"""
        kwargs['order'] = self.pk
        return ReturnOrderLineItem.create(self._api, data=kwargs)

    def getExtraLineItems(self, **kwargs):
        """Return the extra line items associated with this order"""
        return ReturnOrderExtraLineItem.list(self._api, order=self.pk, **kwargs)

    def addExtraLineItem(self, **kwargs):
        """Create (and return) a new ReturnOrderExtraLineItem against this order"""
        kwargs['order'] = self.pk
        return ReturnOrderExtraLineItem.create(self._api, data=kwargs)

    def issue(self, **kwargs):
        """Issue (send) this order"""
        return self._statusupdate(status='issue', **kwargs)

    def hold(self, **kwargs):
        """Place this order on hold"""
        return self._statusupdate(status='hold', **kwargs)

    def cancel(self, **kwargs):
        """Cancel this order"""
        return self._statusupdate(status='cancel', **kwargs)

    def complete(self, **kwargs):
        """Mark this order as complete"""
        return self._statusupdate(status='complete', **kwargs)


class ReturnOrderLineItem(inventree.base.InventreeObject):
    """Class representing the ReturnOrderLineItem model"""

    URL = 'order/ro-line/'
    MIN_API_VERSION = 104

    def getOrder(self):
        """Return the ReturnOrder to which this ReturnOrderLineItem belongs"""
        return ReturnOrder(self._api, self.order)

    def getStockItem(self):
        """Return the StockItem associated with this line item"""
        return inventree.stock.StockItem(self._api, self.item)


class ReturnOrderExtraLineItem(inventree.base.InventreeObject):
    """Class representing the ReturnOrderExtraLineItem model"""

    URL = 'order/ro-extra-line/'
    MIN_API_VERSION = 104

    def getOrder(self):
        """Return the ReturnOrder to which this line item belongs"""
        return ReturnOrder(self._api, self.order)



# -*- coding: utf-8 -*-

import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.plugin import InvenTreePlugin  # noqa: E402


class PluginTest(InvenTreeTestCase):
    """Unit tests for plugin functionality."""

    def test_plugin_lookup(self):
        """Test plugin lookup by key."""

        if self.api.api_version < InvenTreePlugin.MIN_API_VERSION:
            return

        plugins = InvenTreePlugin.list(self.api)

        self.assertGreater(len(plugins), 0)

        p1 = plugins[0]

        # Access the plugin via primary key value
        p2 = InvenTreePlugin(self.api, p1.key)

        self.assertEqual(p1.key, p2.key)
        self.assertEqual(p1.name, p2.name)

        self.assertEqual(p1._data['pk'], p2._data['pk'])

    def test_plugin_list(self):
        """Test plugin list API."""

        if self.api.api_version < InvenTreePlugin.MIN_API_VERSION:
            return

        plugins = InvenTreePlugin.list(self.api)

        expected_attributes = [
            'pk',
            'key',
            'name',
            'package_name',
            'active',
            'meta',
            'mixins',
            'is_builtin',
            'is_sample',
            'is_installed'
        ]

        for plugin in plugins:
            for key in expected_attributes:
                self.assertIn(key, plugin)

    def test_filter_by_active(self):
        """Filter by plugin active status."""

        if self.api.api_version < InvenTreePlugin.MIN_API_VERSION:
            return

        plugins = InvenTreePlugin.list(self.api, active=True)
        self.assertGreater(len(plugins), 0)

        plugin = plugins[0]

        for plugin in plugins:
            self.assertTrue(plugin.active)

    def test_filter_by_builtin(self):
        """Filter by plugin builtin status."""

        if self.api.api_version < InvenTreePlugin.MIN_API_VERSION:
            return

        plugins = InvenTreePlugin.list(self.api, builtin=True)
        self.assertGreater(len(plugins), 0)

        for plugin in plugins:
            self.assertTrue(plugin.is_builtin)
    
    def test_filter_by_mixin(self):
        """Test that we can filter by 'mixin' attribute."""

        if self.api.api_version < InvenTreePlugin.MIN_API_VERSION:
            return

        n = InvenTreePlugin.count(self.api)

        plugins = InvenTreePlugin.list(self.api, mixin='labels')

        self.assertLess(len(plugins), n)
        self.assertGreater(len(plugins), 0)

        for plugin in plugins:
            self.assertIn('labels', plugin.mixins)



# -*- coding: utf-8 -*-

import inventree.base
import inventree.report
import inventree.stock


class Build(
    inventree.base.AttachmentMixin,
    inventree.base.ParameterMixin,
    inventree.base.StatusMixin,
    inventree.base.MetadataMixin,
    inventree.report.ReportPrintingMixin,
    inventree.base.InventreeObject,
):
    """ Class representing the Build database model """

    URL = 'build/'
    MODEL_TYPE = 'build'

    def issue(self):
        """Mark this build as 'issued'."""
        return self._statusupdate(status='issue')

    def hold(self):
        """Mark this build as 'on hold'."""
        return self._statusupdate(status='hold')

    def complete(
        self,
        accept_overallocated='reject',
        accept_unallocated=False,
        accept_incomplete=False,
    ):
        """Finish a build order. Takes the following flags:
        - accept_overallocated
        - accept_unallocated
        - accept_incomplete
        """
        return self._statusupdate(
            status='finish',
            data={
                'accept_overallocated': accept_overallocated,
                'accept_unallocated': accept_unallocated,
                'accept_incomplete': accept_incomplete,
            }
        )

    def finish(self, *args, **kwargs):
        """Alias for complete"""
        return self.complete(*args, **kwargs)

    def getLines(self, **kwargs):
        """ Return the build line items associated with this build order """
        return BuildLine.list(self._api, build=self.pk, **kwargs)

    def getBuildOutputs(self, complete: bool = None, **kwargs):
        """ Return the build output items associated with this build order

        Arguments:
            - complete: If not None, filter the build outputs by their 'complete' status
        """
        if complete is not None:
            kwargs['is_building'] = not complete

        # Find stock items which are marked as 'outputs' of this build order
        return inventree.stock.StockItem.list(
            self._api,
            build=self.pk,
            **kwargs
        )

    def createBuildOutput(self, **kwargs):
        """ Create a new build output (stock item) associated with this build order """
        result = self._api.post(
            f'{self.URL}{self.pk}/create-output/',
            data={
                **kwargs
            }
        )

        # Note: The response is a list of created stock items
        return [inventree.stock.StockItem(self._api, item['pk'], item) for item in result]

    def cancelBuildOutputs(self, outputs):
        """ Cancel a build output item associated with this build order

        Arguments:
            - outputs: The StockItem object (or list of StockItem objects) to cancel
        """

        if not isinstance(outputs, list):
            outputs = [outputs]

        return self._api.post(
            f'{self.URL}{self.pk}/delete-outputs/',
            data={
                'outputs': [
                    {'output': output.pk} for output in outputs
                ]
            }
        )

    def scrapBuildOutput(self, output, **kwargs):
        """ Scrap a single build output item associated with this build order

        Arguments:
            - output: The StockItem object to scrap
        """

        data = {
            **kwargs,
            'outputs': [
                {
                    'output': output.pk,
                    'quantity': kwargs.get('quantity', output.quantity),
                }
            ]
        }

        data['location'] = kwargs.get('location', output.location)

        return self._api.post(
            f'{self.URL}{self.pk}/scrap-outputs/',
            data=data
        )

    def completeBuildOutput(self, output, **kwargs):
        """ Mark a single build output item as complete

        Arguments:
            - output: The StockItem object to mark as complete
        """

        data = {
            **kwargs,
            'outputs': [
                {
                    'output': output.pk,
                    'quantity': kwargs.get('quantity', output.quantity),
                }
            ]
        }

        # If a location is not specified, use the current location of the stock item
        data['location'] = kwargs.get('location', output.location)

        return self._api.post(
            f'{self.URL}{self.pk}/complete/',
            data=data
        )


class BuildLine(
    inventree.base.InventreeObject,
):
    """ Class representing the BuildLine database model """

    URL = 'build/line/'
    MODEL_TYPE = 'buildline'

    def getBuild(self):
        """Return the Build object associated with this line item"""
        return Build(self._api, self.build)


class BuildItem(
    inventree.base.InventreeObject,
):
    """ Class representing the BuildItem database model """

    URL = 'build/item/'
    MODEL_TYPE = 'builditem'

    def getBuild(self):
        """Return the Build object associated with this build item"""
        return Build(self._api, self.build)

    def getBuildLine(self):
        """Return the BuildLine object associated with this build item"""
        return BuildLine(self._api, self.build_line)

    def getStockItem(self):
        """Return the StockItem object associated with this build item"""
        return inventree.stock.StockItem(self._api, self.stock_item)



# -*- coding: utf-8 -*-

import json
import logging
import requests
import os

from . import api as inventree_api

INVENTREE_PYTHON_VERSION = "0.23.0"


logger = logging.getLogger('inventree')


class InventreeObject(object):
    """ Base class for an InvenTree object """

    # API URL (required) for the particular model type
    URL = ""

    @classmethod
    def get_url(cls):
        """Helper method to get the URL associated with this model."""

        url = cls.URL

        if not url.endswith('/'):
            url += '/'

        return url

    # Minimum server version for the particular model type
    MIN_API_VERSION = None
    MAX_API_VERSION = None

    MODEL_TYPE = None

    @classmethod
    def getPkField(cls):
        """Return the primary key field name for this model.

        The default value (used for most models) is 'pk'.
        """
        return 'pk'

    def getPkValue(self):
        """Return the primary key value for this model."""

        return self._data.get(self.getPkField(), None)

    @property
    def pk(self):
        """Override the 'pk' property to return the primary key value for this object.

        Note that by default this is the 'pk' field, but can be overridden in subclasses.
        """
        val = self.getPkValue()

        # Coerce 'pk' values to integer
        if self.getPkField() == 'pk':
            val = int(val)
        
        return val

    def __str__(self):
        """
        Simple human-readable printing.
        Can override in subclass
        """

        return f"{type(self)}<{self.getPkField()}={self.pk}>"

    def __init__(self, api, pk=None, data=None):
        """ Instantiate this InvenTree object.

        Args:
            api - The request manager object
            pk - The ID (primary key) associated with this object on the server
            data - JSON representation of the object
        """

        self.checkApiVersion(api)

        # If the pk is not explicitly provided,
        # extract it from the provided dataset
        if pk is None and data:
            pk = data.get(self.getPkField(), None)

        if self.getPkField() == 'pk' and pk is not None:
            try:
                pk = int(str(pk).strip())
            except Exception:
                raise TypeError(f"Invalid primary key value '{pk}' for {self.__class__}")
            
            if pk <= 0:
                raise ValueError(f"Supplier <pk> value ({pk}) for {self.__class__} must be positive.")

        url = self.get_url()

        self._url = f"{url}/{pk}/"
        self._api = api

        if data is None:
            data = {}

        self._data = data

        # If the data are not populated, fetch from server
        if len(self._data) == 0:
            self.reload()

    @classmethod
    def getModelType(cls):
        """Return the model type for this label printing class."""
        return cls.MODEL_TYPE

    @classmethod
    def checkApiVersion(cls, api):
        """Check if the API version supports this particular model.

        Raises:
            NotSupportedError if the server API version is too 'old'
        """

        if cls.MIN_API_VERSION and cls.MIN_API_VERSION > api.api_version:
            raise NotImplementedError(f"Server API Version ({api.api_version}) is too old for the '{cls.__name__}' class, which requires API version >= {cls.MIN_API_VERSION}")

        if cls.MAX_API_VERSION and cls.MAX_API_VERSION < api.api_version:
            raise NotImplementedError(f"Server API Version ({api.api_version}) is too new for the '{cls.__name__}' class, which requires API version <= {cls.MAX_API_VERSION}")

    @classmethod
    def options(cls, api):
        """Perform an OPTIONS request for this model, to determine model information.

        InvenTree provides custom metadata for each API endpoint, accessed via a HTTP OPTIONS request.
        This endpoint provides information on the various fields available for that endpoint.
        """

        cls.checkApiVersion(api)

        response = api.request(
            cls.get_url(),
            method='OPTIONS',
        )

        if not response.status_code == 200:
            logger.error(f"OPTIONS for '{cls.URL}' returned code {response.status_code}")
            return {}

        try:
            data = json.loads(response.text)
        except json.decoder.JSONDecodeError:
            logger.error(f"Error decoding OPTIONS response for '{cls.URL}'")
            return {}

        return data

    @classmethod
    def fields(cls, api):
        """
        Returns a list of available fields for this model.

        Introspects the available fields using an OPTIONS request.
        """

        opts = cls.options(api)

        actions = opts.get('actions', {})
        post = actions.get('POST', {})

        return post

    @classmethod
    def fieldInfo(cls, field_name, api):
        """Return metadata for a specific field on a model"""

        fields = cls.fields(api)

        if field_name in fields:
            return fields[field_name]
        else:
            logger.warning(f"Field '{field_name}' not found in OPTIONS request for {cls.URL}")
            return {}

    @classmethod
    def fieldNames(cls, api):
        """
        Return a list of available field names for this model
        """

        return [k for k in cls.fields(api).keys()]

    @classmethod
    def create(cls, api, data, **kwargs):
        """ Create a new database object in this class. """

        cls.checkApiVersion(api)

        # Ensure the pk value is None so an existing object is not updated
        if cls.getPkField() in data.keys():
            data.pop(cls.getPkField())

        response = api.post(cls.URL, data, **kwargs)

        if response is None:
            logger.error("Error creating new object")
            return None

        return cls(api, data=response)

    @classmethod
    def count(cls, api, **kwargs):
        """Return a count of all items of this class in the database"""

        params = kwargs

        # By limiting to a single result, we perform a fast query, but get a total number of results
        params['limit'] = 1

        response = api.get(url=cls.URL, params=params)

        return response['count']

    @classmethod
    def list(cls, api, **kwargs):
        """Return a list of all items in this class on the database.

        Requires:

        URL - Base URL
        """
        cls.checkApiVersion(api)

        # Check if custom URL is present in request arguments
        if 'url' in kwargs:
            url = kwargs.pop('url')
        else:
            url = cls.URL

        try:
            response = api.get(url=url, params=kwargs)
        except requests.exceptions.HTTPError as e:
            logger.error(f"Error during list request: {e}")
            # Return an empty list

            raise_error = kwargs.get('raise_error', False)

            if raise_error:
                raise e
            else:
                return []

        if response is None:
            return []

        items = []

        if isinstance(response, dict) and response['results'] is not None:
            response = response['results']

        for data in response:
            if cls.getPkField() in data:
                items.append(cls(data=data, api=api))

        return items

    def delete(self):
        """ Delete this object from the database """

        self.checkApiVersion(self._api)

        if self._api:
            return self._api.delete(self._url)

    def save(self, data=None, files=None, method='PATCH'):
        """
        Save this object to the database
        """

        self.checkApiVersion(self._api)

        # If 'data' is not specified, then use *all* the data
        if data is None:
            data = self._data

        if self._api:

            # Default method used is PATCH (partial update)
            if method.lower() == 'patch':
                response = self._api.patch(self._url, data, files=files)
            elif method.lower() == 'put':
                response = self._api.put(self._url, data, files=files)
            else:
                logger.warning(f"save() called with unknown method '{method}'")
                return

        # Automatically re-load data from the returned data
        if response is not None:
            self._data = response
        else:
            self.reload()

        return response

    def is_valid(self):
        """
        Test if this object is 'valid' - it has received data from the server.

        To be considered 'valid':

        - Must have a non-null PK
        - Must have a non-null and non-empty data structure
        """

        data = getattr(self, '_data', None)

        if self.pk is None:
            return False

        if data is None:
            return False

        if len(data) == 0:
            return False

        return True

    def reload(self):
        """ Reload object data from the database """

        self.checkApiVersion(self._api)

        if self._api:
            data = self._api.get(self._url)

            if data is None:
                logger.error(f"Error during reload at {self._url}")
            else:
                self._data = data

            if not self.is_valid():
                logger.error(f"Error during reload at {self._url} - returned data is invalid")

        else:
            raise AttributeError(f"model.reload failed at '{self._url}': No API instance provided")

    def keys(self):
        return self._data.keys()

    def __contains__(self, name):
        return name in self._data

    def __getattr__(self, name):
        try:
            data = object.__getattribute__(self, "_data")
        except AttributeError:
            # Appears to happen during pickling. Raise immediately to prevent recursion errors
            raise AttributeError(name)

        if name in data:
            return data[name]

        # if we're in this block, there already wasn't a "normal" attribute with this name. Raise
        raise AttributeError(name)

    def __getitem__(self, name):
        if name in self._data.keys():
            return self._data[name]
        else:
            raise KeyError(f"Key '{name}' does not exist in dataset")

    def __setitem__(self, name, value):
        if name in self._data.keys():
            self._data[name] = value
        else:
            raise KeyError(f"Key '{name}' does not exist in dataset")


class BulkDeleteMixin:
    """Mixin class for models which support 'bulk deletion'

    - Perform a DELETE operation against the LIST endpoint for the model
    - Provide a list of items to be deleted, or filters to apply

    Requires API version 58
    """

    @classmethod
    def bulkDelete(cls, api: inventree_api.InvenTreeAPI, items=None, filters=None):
        """Perform bulk delete operation

        Arguments:
            api: InventreeAPI instance
            items: Optional list of items (pk values) to be deleted
            filters: Optional query filters to delete

        Returns:
            API response object

        Throws:
            NotImplementError: The server API version is too old (requires v58)
            ValueError: Neither items or filters are supplied

        """

        if not items and not filters:
            raise ValueError("Must supply either 'items' or 'filters' argument")

        data = {}

        if items:
            data['items'] = items

        if filters:
            data['filters'] = filters

        return api.delete(
            cls.URL,
            json=data,
        )


class Attachment(BulkDeleteMixin, InventreeObject):
    """Class representing a file attachment object."""

    URL = 'attachment/'

    # Ref: https://github.com/inventree/InvenTree/pull/7420
    MIN_API_VERSION = 207

    @classmethod
    def add_link(cls, api, link, comment="", **kwargs):
        """
        Add an external link attachment.

        Args:
            api: Authenticated InvenTree API instance
            link: External link to attach
            comment: Add comment to the attachment
            kwargs: Additional kwargs to suppl
        """

        data = kwargs
        data["comment"] = comment
        data["link"] = link

        if response := api.post(cls.URL, data):
            logger.info(f"Link attachment added to {cls.URL}")
        else:
            logger.error(f"Link attachment failed at {cls.URL}")

        return response

    @classmethod
    def upload(cls, api, attachment, comment='', **kwargs):
        """
        Upload a file attachment.
        Ref: https://2.python-requests.org/en/master/user/quickstart/#post-a-multipart-encoded-file

        Args:
            api: Authenticated InvenTree API instance
            attachment: Either a file object, or a filename (string)
            comment: Add comment to the upload
            kwargs: Additional kwargs to supply
        """

        data = kwargs
        data['comment'] = comment

        if type(attachment) is str:
            if not os.path.exists(attachment):
                raise FileNotFoundError(f"Attachment file '{attachment}' does not exist")

            # Load the file as an in-memory file object
            with open(attachment, 'rb') as fo:
                response = api.post(
                    Attachment.URL,
                    data,
                    files={
                        'attachment': (os.path.basename(attachment), fo),
                    }
                )

        else:
            # Assumes a StringIO or BytesIO like object
            name = getattr(attachment, 'name', 'filename')

            response = api.post(
                Attachment.URL,
                data,
                files={
                    'attachment': (name, attachment),
                }
            )

        if response:
            logger.info(f"File uploaded to {cls.URL}")
        else:
            logger.error(f"File upload failed at {cls.URL}")

        return response

    def download(self, destination, **kwargs):
        """
        Download the attachment file to the specified location
        """

        return self._api.downloadFile(self.attachment, destination, **kwargs)


class AttachmentMixin:
    """Mixin class which allows a model class to interact with attachments."""

    def getAttachments(self, **kwargs):
        """Return a list of attachments associated with this object."""

        return Attachment.list(
            self._api,
            model_type=self.getModelType(),
            model_id=self.pk,
            **kwargs
        )

    def uploadAttachment(self, attachment, comment=""):
        """Upload a file attachment against this model instance."""

        return Attachment.upload(
            self._api,
            attachment,
            comment=comment,
            model_type=self.getModelType(),
            model_id=self.pk
        )
        
    def addLinkAttachment(self, link, comment=""):
        """Add an external link attachment against this Object.

        Args:
            link: The link to attach
            comment: Attachment comment
        """

        return Attachment.add_link(
            self._api,
            link,
            comment=comment,
            model_type=self.getModelType(),
            model_id=self.pk
        )


class Parameter(BulkDeleteMixin, InventreeObject):
    """Class representing a custom parameter object."""

    URL = "parameter/"

    # Ref: https://github.com/inventree/InvenTree/pull/10699
    MIN_API_VERSION = 429


class ParameterTemplate(InventreeObject):
    """Class representing a parameter template object."""

    URL = "parameter/template/"

    # Ref: https://github.com/inventree/InvenTree/pull/10699
    MIN_API_VERSION = 429


class ParameterMixin:
    """Mixin class which allows a model class to interact with parameters.
    
    Ref: https://github.com/inventree/InvenTree/pull/10699
    """

    def getParameters(self, **kwargs):
        """Return a list of parameters associated with this object."""

        if self._api.api_version < Parameter.MIN_API_VERSION:
            raise NotImplementedError(f"Server API Version ({self._api.api_version}) is too old for ParameterMixin, which requires API version >= {Parameter.MIN_API_VERSION}")

        return Parameter.list(
            self._api,
            model_type=self.getModelType(),
            model_id=self.pk,
            **kwargs
        )


class MetadataMixin:
    """Mixin class for models which support a 'metadata' attribute.

    - The 'metadata' is not used for any InvenTree business logic
    - Instead it can be used by plugins for storing arbitrary information
    - Internally it is stored as a JSON database field
    - Metadata is accessed via the API by appending './metadata/' to the API URL

    Note: Requires server API version 49 or newer

    """
    NEW_METADATA_API_VERSION = 436

    @property
    def metadata_url(self):
        """Return the metadata URL for this model instance."""
        if self._api.api_version < self.NEW_METADATA_API_VERSION:
            return os.path.join(self._url, "metadata/")
        model_type = self.getModelType()
        return f"metadata/{model_type}/{self.pk}/"

    def getMetadata(self):
        """Read model instance metadata"""
        if self._api:

            response = self._api.get(self.metadata_url)

            return response['metadata']
        else:
            raise AttributeError(f"model.getMetadata failed at '{self._url}': No API instance provided")

    def setMetadata(self, data, overwrite=False):
        """Write metadata to this particular model.

        Arguments:
            data: The data to be written. Must be a dict object
            overwrite: If true, provided data replaces existing data. If false (default) data is merged with any existing data.
        """

        if type(data) is not dict:
            raise TypeError("Data provided to 'setMetadata' method must be a dict object")

        if self._api:

            if overwrite:
                return self._api.put(
                    self.metadata_url,
                    data={
                        "metadata": data,
                    }
                )
            else:
                return self._api.patch(
                    self.metadata_url,
                    data={
                        "metadata": data
                    }
                )
        else:
            raise AttributeError(f"model.setMetadata failed at '{self._url}': No API instance provided")


class ImageMixin:
    """Mixin class for supporting image upload against a model.

    - The model must have a specific 'image' field associated
    """

    def uploadImage(self, image):
        """
        Upload an image against this model.

        Args:
            image: Either an image file (BytesIO) or a filename path
        """

        files = {}

        # string image = filename
        if type(image) is str:
            if os.path.exists(image):
                f = os.path.basename(image)

                with open(image, 'rb') as fo:
                    files['image'] = (f, fo)

                    return self.save(
                        data={},
                        files=files
                    )
            else:
                raise FileNotFoundError(f"Image file does not exist: '{image}'")

        # TODO: Support upload of in-memory images (e.g. Image / BytesIO)

        else:
            raise TypeError(f"uploadImage called with invalid image: '{image}'")

    def downloadImage(self, destination, **kwargs):
        """
        Download the image for this Part, to the specified destination
        """

        if self.image:
            return self._api.downloadFile(self.image, destination, **kwargs)
        else:
            raise ValueError(f"Part '{self.name}' does not have an associated image")


class StatusMixin:
    """Class adding functionality to assign a new status by calling
    - complete
    - cancel
    on supported items.

    Other functions, such as
    - ship
    - finish
    - issue
    can be reached through _statusupdate function
    """

    def _statusupdate(self, status: str, reload=True, data=None, **kwargs):

        # Check status
        if status not in [
            'complete',
            'cancel',
            'hold',
            'ship',
            'issue',
            'finish',
        ]:
            raise ValueError(f"Order stats {status} not supported.")

        # Set the url
        URL = self.URL + f"/{self.pk}/{status}/"

        if data is None:
            data = {}

        data.update(kwargs)

        # Send data
        response = self._api.post(URL, data)

        # Reload
        if reload:
            self.reload()

        # Return
        return response

    def complete(self, **kwargs):

        return self._statusupdate(status='complete', **kwargs)

    def cancel(self, **kwargs):

        return self._statusupdate(status='cancel', **kwargs)


class BarcodeMixin:
    """Adds barcode scanning functionality to various data types.

    Any class which inherits from this mixin can assign (or un-assign) barcode data.
    """

    @classmethod
    def barcodeModelType(cls):
        """Return the model type name required for barcode assignment.

        Default value is the lower-case class name ()
        """
        return cls.__name__.lower()

    def assignBarcode(self, barcode_data: str, reload=True):
        """Assign an arbitrary barcode to this object (in the database).

        Arguments:
            barcode_data: A string containing arbitrary barcode data
        """

        model_type = self.barcodeModelType()

        response = self._api.post(
            'barcode/link/',
            {
                'barcode': barcode_data,
                model_type: self.pk,
            }
        )

        if reload:
            self.reload()

        return response

    def unassignBarcode(self, reload: bool = True):
        """Unassign a barcode from this object"""

        model_type = self.barcodeModelType()

        response = self._api.post(
            'barcode/unlink/',
            {
                model_type: self.pk,
            }
        )

        if reload:
            self.reload()

        return response



# -*- coding: utf-8 -*-

import os
import sys
import unittest

import requests

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from inventree import api  # noqa: E402
from inventree import base  # noqa: E402
from inventree import part  # noqa: E402
from inventree import stock  # noqa: E402

SERVER = os.environ.get('INVENTREE_PYTHON_TEST_SERVER', 'http://127.0.0.1:12345')
USERNAME = os.environ.get('INVENTREE_PYTHON_TEST_USERNAME', 'testuser')
PASSWORD = os.environ.get('INVENTREE_PYTHON_TEST_PASSWORD', 'testpassword')


class URLTests(unittest.TestCase):
    """Class for testing URL functionality"""

    def test_base_url(self):
        """Test validation of URL provided to InvenTreeAPI class"""

        # Each of these URLs should be invalid
        for url in [
            "test.com/123",
            "http://:80/123",
            "//xyz.co.uk",
        ]:
            with self.assertRaises(Exception):
                a = api.InvenTreeAPI(url, connect=False)

        # test for base URL construction
        a = api.InvenTreeAPI('https://test.com', connect=False)
        self.assertEqual(a.base_url, 'https://test.com/')
        self.assertEqual(a.api_url, 'https://test.com/api/')

        # more tests that the base URL is set correctly under specific conditions
        urls = [
            "http://a.b.co:80/sub/dir/api/",
            "http://a.b.co:80/sub/dir/api",
            "http://a.b.co:80/sub/dir/",
            "http://a.b.co:80/sub/dir",
        ]

        for url in urls:
            a = api.InvenTreeAPI(url, connect=False)
            self.assertEqual(a.base_url, "http://a.b.co:80/sub/dir/")
            self.assertEqual(a.api_url, "http://a.b.co:80/sub/dir/api/")

    def test_url_construction(self):
        """Test that the API class correctly constructs URLs"""

        a = api.InvenTreeAPI("http://localhost:1234", connect=False)

        tests = {
            'part/': 'http://localhost:1234/api/part/',
            '/api/stock/': 'http://localhost:1234/api/stock/',
            '/plugin/part/': 'http://localhost:1234/plugin/part/',
            'order/so/shipment': 'http://localhost:1234/api/order/so/shipment',
            'https://example.com/': 'https://example.com/',
        }

        for endpoint, url in tests.items():
            self.assertEqual(a.constructApiUrl(endpoint), url)


class LoginTests(unittest.TestCase):

    def test_failed_logins(self):

        # Attempt connection where no server exists
        with self.assertRaises(Exception):
            a = api.InvenTreeAPI("http://127.0.0.1:9999", username="admin", password="password")

        # Attempt connection with invalid credentials
        with self.assertRaises(Exception):
            a = api.InvenTreeAPI(SERVER, username="abcde", password="********")

            self.assertIsNotNone(a.server_details)
            self.assertIsNone(a.token)


class Unauthenticated(unittest.TestCase):
    """
    Test that we cannot access the data if we are not authenticated.
    """

    def setUp(self):
        self.api = api.InvenTreeAPI(SERVER, username="hello", password="world", connect=False)

    def test_read_parts(self):

        with self.assertRaises(Exception) as ar:
            part.Part.list(self.api)

        self.assertIn('Authentication at InvenTree server failed', str(ar.exception))

    def test_file_download(self):
        """
        Attempting to download a file while unauthenticated should raise an error
        """

        # Downloading without auth = unauthorized error (401)
        with self.assertRaises(requests.exceptions.HTTPError):
            self.assertFalse(self.api.downloadFile('/media/part/files/1/test.pdf', 'test.pdf'))


class Timeout(unittest.TestCase):
    """
    Test that short timeout leads to correct error
    """

    def test_timeout(self):
        """
        This unrealistically short timeout should lead to a timeout error
        """
        # Attempt connection with short timeout
        with self.assertRaises(requests.exceptions.ReadTimeout):
            a = api.InvenTreeAPI(SERVER, username=USERNAME, password=PASSWORD, timeout=0.001)  # noqa: F841


class InvenTreeTestCase(unittest.TestCase):
    """
    Base class for running InvenTree unit tests.

    - Creates an authenticated API instance
    """

    def setUp(self):
        """
        Test case setup functions
        """
        self.api = api.InvenTreeAPI(
            SERVER,
            username=USERNAME, password=PASSWORD,
            timeout=30,
            token_name='python-test',
            use_token_auth=True
        )


class InvenTreeAPITest(InvenTreeTestCase):

    def test_token(self):
        self.assertIsNotNone(self.api.token)

    def test_details(self):
        self.assertIsNotNone(self.api.server_details)

        details = self.api.server_details
        self.assertIn('server', details)
        self.assertIn('instance', details)

        self.assertIn('apiVersion', details)

        api_version = int(details['apiVersion'])

        self.assertTrue(api_version >= self.api.getMinApiVersion())


class TestCreate(InvenTreeTestCase):
    """
    Test that objects can be created via the API
    """

    def test_create_stuff(self):

        with self.assertRaises(requests.exceptions.ReadTimeout):
            # Test short timeout for a specific function
            c = part.PartCategory.create(self.api, {
                'parent': None,
                'name': 'My custom category',
                'description': 'A part category',
            }, timeout=0.001)

        n = part.PartCategory.count(self.api)

        # Create a custom category
        c = part.PartCategory.create(self.api, {
            'parent': None,
            'name': f'Custom category {n + 1}',
            'description': 'A part category',
        })

        self.assertIsNotNone(c)
        self.assertIsNotNone(c.pk)

        n = part.Part.count(self.api)

        p = part.Part.create(self.api, {
            'name': f'ACME Widget {n}',
            'description': 'A simple widget created via the API',
            'category': c.pk,
            'ipn': f'ACME-0001-{n}',
            'virtual': False,
            'active': True
        })

        self.assertIsNotNone(p)
        self.assertEqual(p.category, c.pk)

        cat = p.getCategory()
        self.assertEqual(cat.pk, c.pk)

        s = stock.StockItem.create(self.api, {
            'part': p.pk,
            'quantity': 45,
            'notes': 'This is a note',
        })[0]

        self.assertIsNotNone(s)
        self.assertEqual(s.part, p.pk)

        prt = s.getPart()
        self.assertEqual(prt.pk, p.pk)
        self.assertEqual(prt.name, f'ACME Widget {n}')


class TemplateTest(InvenTreeTestCase):

    def test_get_widget(self):

        widget = part.Part(self.api, 10000)

        self.assertEqual(widget.name, "Chair Template")

        test_templates = widget.getTestTemplates()
        self.assertGreaterEqual(len(test_templates), 5)

        keys = [test.key for test in test_templates]

        self.assertIn('teststrengthofchair', keys)
        self.assertIn('applypaint', keys)
        self.assertIn('attachlegs', keys)

    def test_add_template(self):
        """
        Test that we can add a test template via the API
        """

        widget = part.Part(self.api, pk=10000)

        n = len(widget.getTestTemplates())

        part.PartTestTemplate.create(self.api, {
            'part': widget.pk,
            'test_name': f"Test_Name_{n}",
            'description': 'A test or something',
            'required': True,
        })

        self.assertEqual(len(widget.getTestTemplates()), n + 1)

    def test_add_result(self):

        # Look for a particular serial number
        items = stock.StockItem.list(self.api, serial=1000)
        self.assertEqual(len(items), 1)

        item = items[0]

        tests = item.getTestResults()

        n = len(tests)

        # Upload a test result against 'firmware' (should fail the first time)
        args = {
            'attachment': 'test/attachment.txt',
            'value': '0x123456',
        }

        result = item.uploadTestResult('firmwareversion', False, **args)

        self.assertTrue(result)

        item.uploadTestResult('temperaturetest', True)
        item.uploadTestResult('settingschecksum', False, value='some data')

        # There should be 3 more test results now!
        results = item.getTestResults()
        self.assertEqual(len(results), n + 3)


if __name__ == '__main__':
    print("Running InvenTree Python Unit Tests: Version " + base.INVENTREE_PYTHON_VERSION)
    unittest.main()



# -*- coding: utf-8 -*-

import os
import sys

import requests

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree import company  # noqa: E402
from inventree import part  # noqa: E402
from inventree.stock import StockItem, StockLocation  # noqa: E402


class StockLocationTest(InvenTreeTestCase):
    """
    Tests for the StockLocation model

    Fixture data can be found in the InvenTree source:

    - InvenTree/stock/fixtures/location.yaml

    """

    def test_location_list(self):
        """
        Test the LIST API endpoint for the StockLocation model
        """

        locs = StockLocation.list(self.api)
        self.assertGreaterEqual(len(locs), 4)

        for loc in locs:
            self.assertEqual(type(loc), StockLocation)

    def test_location_create(self):
        """
        Check that we can create a new stock location via the APi
        """

        n = len(StockLocation.list(self.api))

        parent = StockLocation(self.api, pk=7)

        n_childs = len(parent.getChildLocations())

        # Create a sublocation with a unique name
        location = StockLocation.create(
            self.api,
            {
                "name": f"My special location {n_childs}",
                "description": "A location created with the API!",
                "parent": 7
            }
        )

        self.assertIsNotNone(location)

        # Now, request back via the API using a secondary object
        loc = StockLocation(self.api, pk=location.pk)

        self.assertEqual(loc.name, f"My special location {n_childs}")
        self.assertEqual(loc.parent, 7)

        # Change the name of the location
        loc.save({
            "name": f"A whole new name {n_childs}",
        })

        # Reload the original object
        location.reload()

        self.assertEqual(location.name, f"A whole new name {n_childs}")

        # Check that the number of locations has been updated
        locs = StockLocation.list(self.api)
        self.assertEqual(len(locs), n + 1)

    def test_location_stock(self):
        """Query stock by location"""
        location = StockLocation(self.api, pk=4)

        self.assertEqual(location.pk, 4)
        self.assertEqual(location.description, "Place of work")

        items = location.getStockItems()

        self.assertGreaterEqual(len(items), 15)

        # Check specific part stock in location 1 (initially empty)
        items = location.getStockItems(part=1)

        n = len(items)

        for i in range(5):
            StockItem.create(
                self.api,
                {
                    "part": 1,
                    "quantity": (i + 1) * 50,
                    "location": location.pk,
                }
            )

            items = location.getStockItems(part=1)
            self.assertEqual(len(items), n + i + 1)

        items = location.getStockItems(part=5)
        self.assertGreaterEqual(len(items), 1)

        for item in items:
            self.assertEqual(item.location, location.pk)
            self.assertEqual(item.part, 5)

    def test_location_parent(self):
        """
        Return the Parent location
        """

        # This location does not have a parent
        location = StockLocation(self.api, pk=4)
        self.assertIsNone(location.parent)
        parent = location.getParentLocation()

        self.assertIsNone(parent)

        # Now, get a location which *does* have a parent
        location = StockLocation(self.api, pk=7)
        self.assertIsNotNone(location.parent)
        self.assertEqual(location.parent, 4)

        parent = location.getParentLocation()

        self.assertEqual(type(parent), StockLocation)
        self.assertEqual(parent.pk, 4)
        self.assertIsNone(parent.parent)
        self.assertIsNone(parent.getParentLocation())

        children = parent.getChildLocations()
        self.assertGreaterEqual(len(children), 2)

        for child in children:
            self.assertEqual(type(child), StockLocation)
            self.assertEqual(child.parent, parent.pk)


class StockTest(InvenTreeTestCase):
    """
    Test alternative ways of getting StockItem objects.

    Fixture data can be found in the InvenTree source:

    - InvenTree/stock/fixtures/stock.yaml
    """

    def test_stock(self):

        items = StockItem.list(self.api, part=1)

        n = len(items)

        self.assertGreaterEqual(n, 2)

        for item in items:
            self.assertEqual(item.part, 1)

        # Request via the Part instance (results should be the same!)
        items = part.Part(self.api, 1).getStockItems()
        self.assertEqual(len(items), n)

    def test_get_stock_item(self):
        """
        StockItem API tests.

        Refer to fixture data in InvenTree/stock/fixtures/stock.yaml
        """

        # Grab the first available stock item
        item = StockItem.list(self.api, in_stock=True, limit=1)[0]

        # Get the Part reference
        prt = item.getPart()

        self.assertEqual(type(prt), part.Part)

        # Move the item to a known location
        item.transferStock(3)
        item.reload()

        location = item.getLocation()

        self.assertEqual(type(location), StockLocation)
        self.assertEqual(location.pk, 3)
        self.assertEqual(location.name, "Dining Room")

    def test_bulk_delete(self):
        """Test bulk deletion of stock items"""

        # Add some items to location 3
        for i in range(10):
            StockItem.create(self.api, {
                'location': 3,
                'part': 1,
                'quantity': i + 50,
            })

        self.assertTrue(len(StockItem.list(self.api, location=3)) >= 10)

        # Delete *all* items from location 3
        StockItem.bulkDelete(self.api, filters={
            'location': 3
        })

        loc = StockLocation(self.api, pk=3)
        items = loc.getStockItems()
        self.assertEqual(len(items), 0)

    def test_barcode_support(self):
        """Test barcode support for the StockItem model"""

        items = StockItem.list(self.api, limit=10)

        for item in items:
            # Delete any existing barcode
            item.unassignBarcode()

            # Perform lookup based on 'internal' barcode
            response = self.api.scanBarcode(
                {
                    "stockitem": item.pk,
                }
            )

            self.assertEqual(response['stockitem']['pk'], item.pk)
            self.assertEqual(response['plugin'], 'InvenTreeBarcode')

            # Assign a custom barcode to this StockItem
            barcode = f"custom-stock-item-{item.pk}"
            item.assignBarcode(barcode)

            response = self.api.scanBarcode(barcode)

            self.assertEqual(response['stockitem']['pk'], item.pk)
            self.assertEqual(response['plugin'], 'InvenTreeBarcode')
            self.assertEqual(response['barcode_data'], barcode)

            item.unassignBarcode()

    def test_serialized(self):
        """Test serializing multiple objects on create"""

        # Create items with serial numbers
        items = StockItem.create(
            self.api,
            {
                "part": 10004,
                "quantity": 3,
                "serial_numbers": "1005,1006,1007"
            }
        )

        self.assertEqual(3, len(items))

        self.assertEqual('1005', items[0].serial)
        self.assertEqual('1006', items[1].serial)
        self.assertEqual('1007', items[2].serial)

        # Delete the items after the test
        for item in items:
            item.delete()


class StockAdjustTest(InvenTreeTestCase):
    """Unit tests for stock 'adjustment' actions"""

    def test_count(self):
        """Test the 'count' action"""

        # Find the first available stock item
        item = StockItem.list(self.api, in_stock=True, limit=1)[0]

        # Count number of tracking entries
        n_tracking = len(item.getTrackingEntries())

        q = item.quantity

        item.countStock(q + 100)
        item.reload()

        self.assertEqual(item.quantity, q + 100)

        item.countStock(q, notes='Why hello there')
        item.reload()
        self.assertEqual(item.quantity, q)

        # 2 tracking entries should have been added
        self.assertEqual(
            len(item.getTrackingEntries()),
            n_tracking + 2
        )

        # The most recent tracking entry should have a note
        t = item.getTrackingEntries()[0]
        self.assertEqual(t.label, 'Stock counted')

        # Check error conditions
        with self.assertRaises(requests.exceptions.HTTPError):
            item.countStock('not a number')

        with self.assertRaises(requests.exceptions.HTTPError):
            item.countStock(-1)

    def test_add_remove(self):
        """Test the 'add' and 'remove' actions"""

        # Find the first available stock item
        item = StockItem.list(self.api, in_stock=True, limit=1)[0]

        n_tracking = len(item.getTrackingEntries())

        q = item.quantity

        # Add some items
        item.addStock(10)
        item.reload()
        self.assertEqual(item.quantity, q + 10)

        # Remove the items again
        item.removeStock(10)
        item.reload()
        self.assertEqual(item.quantity, q)

        # 2 additional tracking entries should have been added
        self.assertTrue(len(item.getTrackingEntries()) > n_tracking)

        # Test error conditions
        for v in [-1, 'gg', None]:
            with self.assertRaises(requests.exceptions.HTTPError):
                item.addStock(v)
            with self.assertRaises(requests.exceptions.HTTPError):
                item.removeStock(v)

    def test_transfer(self):
        """Unit test for 'transfer' action"""

        item = StockItem(self.api, pk=2)

        n_tracking = len(item.getTrackingEntries())

        # Transfer to a StockLocation instance
        location = StockLocation(self.api, pk=1)

        item.transferStock(location)
        item.reload()
        self.assertEqual(item.location, 1)

        # Transfer with a location ID
        item.transferStock(2)
        item.reload()
        self.assertEqual(item.location, 2)

        # 2 additional tracking entries should have been added
        self.assertTrue(len(item.getTrackingEntries()) > n_tracking)

        # Attempt to transfer to an invalid location
        for loc in [-1, 'qqq', 99999, None]:
            with self.assertRaises(requests.exceptions.HTTPError):
                item.transferStock(loc)

        # Attempt to transfer with an invalid quantity
        for q in [-1, None, 'hhhh']:
            with self.assertRaises(requests.exceptions.HTTPError):
                item.transferStock(loc, quantity=q)

    def test_transfer_multiple(self):
        """Test transfer of *multiple* items"""

        items = StockItem.list(self.api, in_stock=True, location=1)

        self.assertTrue(len(items) > 1)

        # Construct data to send
        data = []

        for item in items:
            data.append({
                'pk': item.pk,
                'quantity': item.quantity,
            })

        # Transfer all items into a new location
        StockItem.transferStockItems(self.api, data, 2)

        for item in items:
            item.reload()
            self.assertEqual(item.location, 2)

        # Transfer back to the original location
        StockItem.transferStockItems(self.api, data, 1)

        for item in items:
            item.reload()
            self.assertEqual(item.location, 1)

            history = item.getTrackingEntries()

            self.assertTrue(len(history) >= 2)
            self.assertEqual(history[0].label, 'Location changed')

    def test_assign_stock(self):
        """Test assigning stock to customer"""

        items = StockItem.list(self.api)
        self.assertTrue(len(items) > 1)

        # Get first Company which is a customer
        customer = company.Company.list(self.api, is_customer=True)[0]

        # Get first part which is salable
        assignpart = part.Part.list(self.api, salable=True)[0]

        # Create stock item which can be assigned
        assignitem = StockItem.create(
            self.api,
            {
                "part": assignpart.pk,
                "quantity": 10,
            }
        )

        # Verify a single result was returned
        self.assertEqual(1, len(assignitem))
        assignitem = assignitem[0]

        # Assign the item
        assignitem.assignStock(customer=customer, notes='Sell on the side')

        # Reload the item
        assignitem.reload()

        # Check the item is assigned
        self.assertTrue(assignitem.customer == customer.pk)

    def test_install_stock(self):
        """Test install and uninstall a stock item from another"""

        items = StockItem.list(self.api, available=True)

        self.assertTrue(len(items) > 1)

        # get a parent and a child part
        parent_part = part.Part.list(
            self.api,
            trackable=True,
            assembly=True,
            has_stock=True
        )[0]
        parent_stock = parent_part.getStockItems()[0]
        child_stock = items[0]
        child_part = child_stock.getPart()

        # make sure the child is in the bom of the parent

        items = parent_part.getBomItems(search=child_part.name)
        if not items:
            part.BomItem.create(
                self.api, {
                    'part': parent_part.pk,
                    'sub_part': child_part.pk,
                    'quantity': 1
                }
            )

        self.assertIsNone(child_stock.belongs_to)

        # Attempt to install with incorrect quantity
        with self.assertRaises(requests.exceptions.HTTPError):
            parent_stock.installStock(child_stock, quantity=child_stock.quantity * 2)

        with self.assertRaises(requests.exceptions.HTTPError):
            parent_stock.installStock(child_stock, quantity=-100)

        # install the *entire* child item into the parent
        parent_stock.installStock(child_stock, quantity=child_stock.quantity)
        child_stock.reload()
        self.assertIsNotNone(child_stock.belongs_to)
        self.assertEqual(child_stock.belongs_to, parent_stock.pk)

        # and uninstall it again
        location = StockLocation.list(self.api)[0]
        child_stock.uninstallStock(location)

        # check if the location is set correctly to confirm the uninstall
        child_stock.reload()
        new_location = child_stock.getLocation()
        self.assertTrue(new_location.pk == location.pk)



# -*- coding: utf-8 -*-

import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.build import Build  # noqa: E402
from inventree.report import ReportTemplate  # noqa: E402


class ReportClassesTest(InvenTreeTestCase):
    """Tests for Report functions models"""

    def test_list_templates(self):
        """List the available report templates."""

        templates = ReportTemplate.list(self.api)

        self.assertGreater(len(templates), 0)

        for template in templates:
            for key in ['name', 'description', 'enabled', 'model_type', 'template']:
                self.assertIn(key, template)

        # disable a template
        templates[0].save(data={'enabled': False})

        templates = ReportTemplate.list(self.api, enabled=False)
        self.assertGreater(len(templates), 0)

        # enable a template
        templates[0].save(data={'enabled': True})

        templates = ReportTemplate.list(self.api, enabled=True)
        self.assertGreater(len(templates), 0)
    
    def test_print_report(self):
        """Test report printing."""

        # Find a build to print
        build = Build.list(self.api, limit=1)[0]

        templates = build.getReportTemplates()
        self.assertGreater(len(templates), 0)

        template = templates[0]

        # Print the report
        response = build.printReport(template)

        for key in ['pk', 'output']:
            self.assertIn(key, response)
        
        self.assertIsNotNone(response['output'])



# -*- coding: utf-8 -*-

import logging

import inventree.base
import inventree.order

logger = logging.getLogger('inventree')


class Contact(inventree.base.InventreeObject):
    """Class representing the Contact model"""

    URL = 'company/contact/'
    MIN_API_VERSION = 104


class Address(inventree.base.InventreeObject):
    """Class representing the Address model"""

    URL = 'company/address/'
    MIN_API_VERSION = 126


class Company(
    inventree.base.AttachmentMixin,
    inventree.base.ParameterMixin,
    inventree.base.ImageMixin,
    inventree.base.MetadataMixin,
    inventree.base.InventreeObject
):
    """ Class representing the Company database model """

    URL = 'company/'
    MODEL_TYPE = "company"

    def getContacts(self, **kwargs):
        """Return contacts associated with this Company"""
        kwargs['company'] = self.pk
        return Contact.list(self._api, **kwargs)

    def getAddresses(self, **kwargs):
        """Return addresses associated with this Company"""
        kwargs['company'] = self.pk
        return Address.list(self._api, **kwargs)

    def getSuppliedParts(self, **kwargs):
        """
        Return list of SupplierPart objects supplied by this Company
        """
        return SupplierPart.list(self._api, supplier=self.pk, **kwargs)

    def getManufacturedParts(self, **kwargs):
        """
        Return list of ManufacturerPart objects manufactured by this Company
        """
        return ManufacturerPart.list(self._api, manufacturer=self.pk, **kwargs)

    def getPurchaseOrders(self, **kwargs):
        """
        Return list of PurchaseOrder objects associated with this company
        """
        return inventree.order.PurchaseOrder.list(self._api, supplier=self.pk, **kwargs)

    def createPurchaseOrder(self, **kwargs):
        """
        Create (and return) a new PurchaseOrder against this company
        """

        kwargs['supplier'] = self.pk

        return inventree.order.PurchaseOrder.create(
            self._api,
            data=kwargs
        )

    def getSalesOrders(self, **kwargs):
        """
        Return list of SalesOrder objects associated with this company
        """
        return inventree.order.SalesOrder.list(self._api, customer=self.pk, **kwargs)

    def createSalesOrder(self, **kwargs):
        """
        Create (and return) a new SalesOrder against this company
        """

        kwargs['customer'] = self.pk

        return inventree.order.SalesOrder.create(
            self._api,
            data=kwargs
        )

    def getReturnOrders(self, **kwargs):
        """Return list of ReturnOrder objects associated with this company"""
        return inventree.order.ReturnOrder.list(self._api, customer=self.pk, **kwargs)

    def createReturnOrder(self, **kwargs):
        """Create (and return) a new ReturnOrder against this company"""
        kwargs['customer'] = self.pk

        return inventree.order.ReturnOrder.create(self._api, data=kwargs)


class SupplierPart(
    inventree.base.AttachmentMixin,
    inventree.base.ParameterMixin,
    inventree.base.BarcodeMixin,
    inventree.base.BulkDeleteMixin,
    inventree.base.MetadataMixin,
    inventree.base.InventreeObject
):
    """Class representing the SupplierPart database model

    - Implements the BulkDeleteMixin
    """

    URL = 'company/part/'

    def getPriceBreaks(self):
        """ Get a list of price break objects for this SupplierPart """

        return SupplierPriceBreak.list(self._api, part=self.pk)


class ManufacturerPart(
    inventree.base.AttachmentMixin,
    inventree.base.ParameterMixin,
    inventree.base.BulkDeleteMixin,
    inventree.base.MetadataMixin,
    inventree.base.InventreeObject,
):
    """Class representing the ManufacturerPart database model

    - Implements the BulkDeleteMixin
    """

    URL = 'company/part/manufacturer/'
    MODEL_TYPE = "manufacturerpart"

    def getParameters(self, **kwargs):
        """
        GET a list of all ManufacturerPartParameter objects for this ManufacturerPart
        """

        # Support legacy API version which uses a different endpoint
        if self._api.api_version < inventree.base.Parameter.MIN_API_VERSION:
            return ManufacturerPartParameter.list(self._api, manufacturer_part=self.pk, **kwargs)

        return super().getParameters(**kwargs)


class ManufacturerPartParameter(inventree.base.BulkDeleteMixin, inventree.base.InventreeObject):
    """Class representing the ManufacturerPartParameter database model.

    Note: This class was removed in API version 418 and later.
    Ref: https://github.com/inventree/InvenTree/pull/10699
    """

    URL = 'company/part/manufacturer/parameter/'
    MAX_API_VERSION = 428


class SupplierPriceBreak(inventree.base.InventreeObject):
    """ Class representing the SupplierPriceBreak database model """

    URL = 'company/price-break/'



# -*- coding: utf-8 -*-

import os
import sys

from requests.exceptions import HTTPError

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.base import Attachment  # noqa: E402
from inventree import company  # noqa: E402
from inventree import order  # noqa: E402
from inventree import part  # noqa: E402
from inventree import stock  # noqa: E402


class POTest(InvenTreeTestCase):
    """
    Unit tests for PurchaseOrder
    """

    def test_po_fields(self):
        """
        Check that the OPTIONS endpoint provides field names for this model
        """

        field_names = order.PurchaseOrder.fieldNames(self.api)

        names = [
            'issue_date',
            'complete_date',
            'supplier',
            'status',
        ]

        for name in names:
            self.assertIn(name, field_names)

    def test_po_create(self):
        """
        Test purchase order creation
        """

        supplier = company.Company(self.api, pk=2)
        self.assertTrue(supplier.is_supplier)

        # Find all purchaseable parts
        parts = part.Part.list(self.api, purchaseable=True)

        for idx, prt in enumerate(parts):

            if idx >= 10:
                break

            # Check that the part is marked as purchaseable
            self.assertTrue(prt.purchaseable)

            # If the supplier does not have a supplier part, create one
            supplier_parts = supplier.getSuppliedParts(part=prt.pk)

            if len(supplier_parts) == 0:

                supplier_part = company.SupplierPart.create(
                    self.api,
                    data={
                        'part': prt.pk,
                        'supplier': supplier.pk,
                        'SKU': f"SKU-{supplier.pk}-{prt.name}",
                        'MPN': f"MPN-{supplier.pk}-{prt.name}",
                    }
                )

                self.assertIsNotNone(supplier_part)
                self.assertIsNotNone(supplier_part.pk)

        parts = supplier.getSuppliedParts()

        # There should be at least *some* supplied parts
        self.assertTrue(len(parts) > 0)

        # Create a purchase order
        n = len(order.PurchaseOrder.list(self.api))

        # Create a PO with unique reference
        ref = f"PO-{n+1}"

        po = supplier.createPurchaseOrder(
            reference=ref,
            description="This is a PO created using the Python interface"
        )

        self.assertIsNotNone(po)
        self.assertIsNotNone(po.pk)

        found = False

        # Ensure that we can find the pk of the new order
        for _order in supplier.getPurchaseOrders():
            if _order.pk == po.pk:
                found = True
                break

        self.assertTrue(found)

        # Should not be any line-items yet!
        items = po.getLineItems()
        self.assertEqual(len(items), 0)

        # Let's add some!
        supplier_parts = supplier.getSuppliedParts()

        for idx, sp in enumerate(supplier_parts):

            if idx == 0:
                with self.assertRaises(HTTPError):
                    line = po.addLineItem(part=sp.pk, quantity=idx)
                continue

            line = po.addLineItem(part=sp.pk, quantity=idx)

            self.assertEqual(line.getOrder().pk, po.pk)

            self.assertIsNotNone(line)

            # Assert that a new line item has been created
            self.assertEqual(len(po.getLineItems()), idx)

            # Check that the supplier part reference is correct
            self.assertEqual(line.getSupplierPart().pk, sp.pk)

            # Check that the base part reference is correct
            self.assertEqual(line.getPart().pk, sp.part)

        for idx, line in enumerate(po.getLineItems()):
            self.assertEqual(line.quantity, idx + 1)
            self.assertEqual(line.received, 0)
            line.delete()

        # Assert length is now one less than before
        self.assertEqual(len(po.getLineItems()), 0)

        # Should not be any extra-line-items yet!
        extraitems = po.getExtraLineItems()
        self.assertEqual(len(extraitems), 0)

        # Let's add some!
        extraline = po.addExtraLineItem(quantity=1, reference="Transport costs", notes="Extra line item added from Python interface", price=10, price_currency="EUR")

        self.assertEqual(extraline.getOrder().pk, po.pk)

        self.assertIsNotNone(extraline)

        # Assert that a new line item has been created
        self.assertEqual(len(po.getExtraLineItems()), 1)

        # Assert that we can delete the item again
        extraline.delete()

        # Now there should be 0 lines left
        self.assertEqual(len(po.getExtraLineItems()), 0)

    def test_order_cancel(self):
        """Test that we can cancel a PurchaseOrder via the API"""

        n = len(order.PurchaseOrder.list(self.api)) + 1
        ref = f"PO-{n}"

        # Create a new PO
        po = order.PurchaseOrder.create(self.api, data={
            'supplier': 1,
            'reference': ref,
            'description': 'Some new order'
        })

        self.assertEqual(po.status, 10)
        self.assertEqual(po.status_text, "Pending")

        # Cancel the order
        po.cancel()

        self.assertEqual(po.status, 40)
        self.assertEqual(po.status_text, "Cancelled")

    def test_order_complete_with_receive(self):
        """Test that we can complete an order via the API, after receiving
        items via API"""

        n = len(order.PurchaseOrder.list(self.api)) + 1
        ref = f"PO-{n}"

        # First, let's create a new PurchaseOrder
        po = order.PurchaseOrder.create(self.api, data={
            'supplier': 1,
            'reference': ref,
            'description': 'A purchase order with items to be received',
        })

        # Get first location
        use_location = stock.StockLocation.list(self.api, limit=1)[0]

        # Add some line items
        for p in company.SupplierPart.list(self.api, supplier=1, limit=5):
            po.addLineItem(
                part=p.pk,
                quantity=10,
                destination=use_location.pk
            )

        # Check that lines have actually been added
        lines = po.getLineItems()
        self.assertTrue(len(lines) > 0)

        # Initial status code is "PENDING"
        self.assertEqual(po.status, 10)
        self.assertEqual(po.status_text, "Pending")

        # Receive lines too early, assert error
        with self.assertRaises(HTTPError):
            po.receiveAll(location=1)

        # Issue the order
        po.issue()
        po.reload()

        self.assertEqual(po.status, 20)
        self.assertEqual(po.status_text, "Placed")

        # Try to complete the order (should fail, as lines have not been received)
        with self.assertRaises(HTTPError):
            po.complete()

        po.reload()

        # Check that order status has *not* changed
        self.assertEqual(po.status, 20)

        # Prepare one line item for special treatment
        po_line_0 = po.getLineItems()[0]

        # Get list of items currently in stock, for comparison later
        stock_items_before = stock.StockItem.list(self.api, supplier_part=po_line_0.part, location=use_location.pk)
        pks_before = [x.pk for x in stock_items_before]

        # Now, receive some of one of the line item with status ATTENTION
        po_line_0.receive(status=50, quantity=5)

        # Get new list of stock items
        stock_items_after = stock.StockItem.list(self.api, supplier_part=po_line_0.part, location=use_location.pk)
        pks_after = [x.pk for x in stock_items_after]

        # make sure a stock item has been added
        self.assertEqual(len(pks_after), len(pks_before) + 1)

        # Get the newly added PK and stock item
        newpk = list(set(pks_after) - set(pks_before))[0]
        new_stock_item = stock.StockItem(self.api, pk=newpk)

        # Check the last stock item for the expected quantity and status
        self.assertEqual(new_stock_item.quantity, 5)
        self.assertEqual(new_stock_item.status, 50)

        # Receive the rest of this item, with defaults
        po_line_0.receive()

        # Receive all line items
        # Use the ID of the location here
        result = po.receiveAll(location=use_location.pk)

        # Check the result returned
        if self.api.api_version < 385:  # Ref: https://github.com/inventree/InvenTree/pull/10174/
            self.assertIsInstance(result, dict)
            self.assertIn('items', result)
            self.assertIn('location', result)
            # Check that all except one line were marked
            self.assertEqual(len(result['items']), len(po.getLineItems()) - 1)
        else:
            self.assertIsInstance(result, list)
            self.assertEqual(len(result), len(po.getLineItems()) - 1)

        # Receive all line items again - make sure answer is None
        # use the StockLocation item here
        result = po.receiveAll(location=use_location)
        self.assertIsNone(result)

        # Complete the order, do not accept any incomplete lines
        po.complete(accept_incomplete=False)
        po.reload()

        # Check that the order is now complete
        self.assertEqual(po.status, 30)

    def test_order_complete(self):
        """Test that we can complete an order via the API, with un-finished items remaining"""

        n = len(order.PurchaseOrder.list(self.api)) + 1
        ref = f"PO-{n}"

        # First, let's create a new PurchaseOrder
        po = order.PurchaseOrder.create(self.api, data={
            'supplier': 1,
            'reference': ref,
            'description': 'A new purchase order',
        })

        # Add some line items
        for p in company.SupplierPart.list(self.api, supplier=1, limit=5):

            po.addLineItem(
                part=p.pk,
                quantity=10,
            )

        # Check that lines have actually been added
        lines = po.getLineItems()
        self.assertTrue(len(lines) > 0)

        # Initial status code is "PENDING"
        self.assertEqual(po.status, 10)
        self.assertEqual(po.status_text, "Pending")

        # Issue the order
        po.issue()
        po.reload()

        self.assertEqual(po.status, 20)
        self.assertEqual(po.status_text, "Placed")

        # Try to complete the order (should fail, as lines have not been received)
        with self.assertRaises(HTTPError):
            po.complete()

        po.reload()

        # Check that order status has *not* changed
        self.assertEqual(po.status, 20)

        # Now, try to complete the order again, accepting incomplete
        po.complete(accept_incomplete=True)
        po.reload()

        # Check that the order is now complete
        self.assertEqual(po.status, 30)

    def test_purchase_order_delete(self):
        """
        Test that we can delete existing purchase orders
        """

        orders = order.PurchaseOrder.list(self.api)

        self.assertTrue(len(orders) > 0)

        for po in orders:
            po.delete()

        orders = order.PurchaseOrder.list(self.api)
        self.assertEqual(len(orders), 0)

    def test_po_attachment(self):
        """
        Test upload of attachment against a PurchaseOrder
        """

        if self.api.api_version < Attachment.MIN_API_VERSION:
            return

        # Ensure we have a least one purchase order to work with
        n = len(order.PurchaseOrder.list(self.api))

        po = order.PurchaseOrder.create(self.api, {
            'supplier': 1,
            'reference': f'PO-{n + 100}',
            'description': 'A new purchase order',
        })

        attachments = po.getAttachments()
        self.assertEqual(len(attachments), 0)

        # Test we can upload an attachment against this PurchaseOrder
        fn = os.path.join(os.path.dirname(__file__), 'attachment.txt')

        # Should be able to upload the same file multiple times!
        for i in range(3):
            response = po.uploadAttachment(fn, comment='Test upload to purchase order')
            self.assertEqual(response['comment'], 'Test upload to purchase order')

        # Test that an invalid file raises an error
        with self.assertRaises(FileNotFoundError):
            po.uploadAttachment('not_found.txt')

        # Check that attachments uploaded OK
        attachments = po.getAttachments()
        self.assertEqual(len(attachments), 3)

    def test_invalid_list(self):
        """Test list with an invalid parameter.
        
        Ref: https://github.com/inventree/inventree-python/issues/246
        """

        from inventree.project_code import ProjectCode

        results = order.PurchaseOrder.list(self.api, project_code=999999999)
        self.assertEqual(len(results), 0)

        # Try the same again, but raise the eror
        with self.assertRaises(HTTPError):
            results = order.PurchaseOrder.list(
                self.api,
                project_code=999999999,
                raise_error=True
            )

        # Find a valid project code
        n = ProjectCode.count(self.api)

        # Create a new project code
        pc = ProjectCode.create(self.api, {
            'code': f"TEST-{n+1}",
            'description': 'Test project code',
        })

        # Attach project code to an order
        po = order.PurchaseOrder.list(self.api, limit=1)[0]
        po.save({'project_code': pc.pk})

        # Now, list orders with the project code
        results = order.PurchaseOrder.list(self.api, project_code=pc.pk)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].pk, po.pk)


class SOTest(InvenTreeTestCase):
    """
    Unit test for SalesOrder
    """

    def test_so_fields(self):
        """
        Check that the OPTIONS endpoint provides field names for this model
        """

        names = [
            'customer',
            'description',
            'overdue',
            'reference',
        ]

        fields = order.SalesOrder.fieldNames(self.api)

        for name in names:
            self.assertIn(name, fields)

    def test_so_create(self):
        """
        Test sales order creation
        """

        # Create some salable parts (if none exist)

        if len(part.Part.list(self.api, salable=True)) == 0:

            for idx in range(10):

                part.Part.create(
                    self.api,
                    {
                        'name': f"SellPart_{idx}",
                        'description': "A saleable part",
                        'salable': True,
                        'category': 1,
                        'active': True,
                        'IPN': f"PART_{idx}",
                        'revision': 'A',
                    }
                )

        # Find all saleable parts
        salable_parts = part.Part.list(self.api, salable=True)

        # For each customer
        for customer in company.Company.list(self.api, is_customer=True):

            self.assertTrue(customer.is_customer)
            self.assertTrue(len(salable_parts) > 0)

            # Create a new sales order for this customer
            n = len(customer.getSalesOrders())

            # Create a new sales order for the company
            sales_order = customer.createSalesOrder(
                description='This is a new SalesOrder!'
            )

            self.assertIsNotNone(sales_order)

            self.assertEqual(len(sales_order.getLineItems()), 0)

            self.assertEqual(len(customer.getSalesOrders()), n + 1)

            # Add a line item for each saleable part
            for idx, p in enumerate(salable_parts):
                line = sales_order.addLineItem(part=p.pk, quantity=idx)

                self.assertEqual(line.quantity, idx)

                self.assertEqual(line.getPart().pk, p.pk)

                self.assertEqual(line.getOrder().pk, sales_order.pk)

                self.assertEqual(len(sales_order.getLineItems()), idx + 1)

            # Should not be any extra-line-items yet!
            self.assertEqual(len(sales_order.getExtraLineItems()), 0)

            # Let's add some!
            extraline = sales_order.addExtraLineItem(
                quantity=1,
                reference="Transport costs",
                notes="Extra line item added from Python interface",
                price=10, price_currency="EUR"
            )

            self.assertIsNotNone(extraline)

            self.assertEqual(extraline.getOrder().pk, sales_order.pk)

            # Assert that a new line item has been created
            self.assertEqual(len(sales_order.getExtraLineItems()), 1)

            # Assert that we can delete the item again
            extraline.delete()

            # Now there should be 0 lines left
            self.assertEqual(len(sales_order.getExtraLineItems()), 0)

    def test_so_attachment(self):
        """
        Test upload of attachment against a SalesOrder
        """

        if self.api.api_version < Attachment.MIN_API_VERSION:
            return

        # Grab the first available SalesOrder
        orders = order.SalesOrder.list(self.api)

        if len(orders) > 0:
            so = orders[0]
        else:
            so = order.SalesOrder.create(self.api, {
                'customer': 4,
                "description": "Selling some stuff",
            })

        n = len(so.getAttachments())

        # Upload a new attachment
        fn = os.path.join(os.path.dirname(__file__), 'attachment.txt')
        response = so.uploadAttachment(fn, comment='Sales order attachment')

        pk = response['pk']

        attachment = Attachment(self.api, pk=pk)

        self.assertEqual(attachment.model_type, 'salesorder')
        self.assertEqual(attachment.model_id, so.pk)
        self.assertEqual(attachment.comment, 'Sales order attachment')

        attachments = so.getAttachments()
        self.assertEqual(len(attachments), n + 1)

    def test_so_shipment(self):
        """Test shipment functionality for a SalesOrder."""

        # Construct a new SalesOrder instance
        so = order.SalesOrder.create(self.api, {
            'customer': 4,
            "description": "Selling some stuff",
        })

        # Add some line items to the SalesOrder
        for p in part.Part.list(self.api, is_template=False, salable=True, limit=5):

            # Create a line item matching the part
            order.SalesOrderLineItem.create(
                self.api,
                data={
                    'part': p.pk,
                    'order': so.pk,
                    'quantity': 10,
                }
            )

            # Ensure there is available stock
            stock.StockItem.create(
                self.api,
                data={
                    'part': p.pk,
                    'quantity': 25,
                    'location': 1,
                }
            )

        # The shipments list should return something which is not none
        self.assertIsNotNone(so.getShipments())

        # Count number of current shipments
        num_shipments = len(so.getShipments())

        # Create a new shipment - without data, use SalesOrderShipment method
        with self.assertRaises(TypeError):
            shipment_1 = order.SalesOrderShipment.create(self.api)

        # Create new shipment - minimal data, use SalesOrderShipment method
        shipment_1 = order.SalesOrderShipment.create(
            self.api, data={
                'order': so.pk,
                'reference': f'Package {num_shipments+1}'
            }
        )

        # Assert the shipment is created
        self.assertIsNotNone(shipment_1)

        # Assert the shipment Order is equal to the expected one
        self.assertEqual(shipment_1.getOrder().pk, so.pk)

        # Count number of current shipments
        self.assertEqual(len(so.getShipments()), num_shipments + 1)
        num_shipments = len(so.getShipments())

        # Create new shipment - use addShipment method.
        # Should fail because reference will not be unique
        with self.assertRaises(HTTPError):
            shipment_2 = so.addShipment(f'Package {num_shipments}')

        # Create new shipment - use addShipment method. No extra data
        shipment_2 = so.addShipment(f'Package {num_shipments+1}')

        # Assert the shipment is not created
        self.assertIsNotNone(shipment_2)

        # Assert the shipment Order is equal to the expected one
        self.assertEqual(shipment_2.getOrder().pk, so.pk)

        # Assert shipment reference is as expected
        self.assertEqual(shipment_2.reference, f'Package {num_shipments+1}')

        # Count number of current shipments
        self.assertEqual(len(so.getShipments()), num_shipments + 1)
        num_shipments = len(so.getShipments())

        # Create another shipment - use addShipment method.
        # With some extra data, including non-sense order
        # (which should be overwritten)
        notes = f'Test shipment number {num_shipments+1} for order {so.pk}'
        tracking_number = '93414134343'

        shipment_2 = so.addShipment(
            reference=f'Package {num_shipments+1}',
            order=10103413,
            notes=notes,
            tracking_number=tracking_number
        )

        # Assert the shipment is created
        self.assertIsNotNone(shipment_2)

        # Assert the shipment Order is equal to the expected one
        self.assertEqual(shipment_2.getOrder().pk, so.pk)

        # Assert shipment reference is as expected
        self.assertEqual(shipment_2.reference, f'Package {num_shipments+1}')

        # Make sure extra data is also as expected
        self.assertEqual(shipment_2.tracking_number, tracking_number)

        # Count number of current shipments
        self.assertEqual(len(so.getShipments()), num_shipments + 1)
        num_shipments = len(so.getShipments())

        # Remember for later test
        allocated_quantities = dict()

        # Assign each line item to this shipment
        for si in so.getLineItems():
            # If there is no stock available, delete this line
            if si.available_stock == 0:
                si.delete()
            else:
                response = si.allocateToShipment(shipment_2)
                # Remember what we are doing for later check
                # a response of None means nothing was allocated
                if response is not None:
                    allocated_quantities[si.pk] = (
                        {x['stock_item']: float(x['quantity']) for x in response['items']}
                    )

        # Check saved values
        for so_part in so.getLineItems():
            if so_part.pk in allocated_quantities:
                if len(allocated_quantities[so_part.pk]) > 0:
                    self.assertEqual(
                        {x['item']: float(x['quantity']) for x in shipment_2.allocations if x['line'] == so_part.pk},
                        allocated_quantities[so_part.pk]
                    )

        # Attempt to complete the shipment, but no items have been allocated
        shipment_2.complete()

        # Make sure date is not None
        self.assertIsNotNone(shipment_2.shipment_date)

        # SalesOrderAllocations are broken prior to server API version 267
        if self.api.api_version < 267:
            return

        # Try to complete this order
        # Ship remaining shipments first
        for shp in so.getShipments():

            allocations = shp.getAllocations()

            # Delete shipment if it has no allocations
            if len(allocations) == 0:
                shp.delete()
                continue
            
            # If the shipment has no date, try to mark it shipped
            if shp.shipment_date is None:
                shp.ship()
        
        so.complete()
        self.assertEqual(so.status, 20)
        self.assertEqual(so.status_text, 'Shipped')

    def test_order_cancel(self):
        """Test cancel sales order"""

        so = order.SalesOrder.create(self.api, {
            'customer': 4,
            "description": "Selling some more stuff",
        })

        so.cancel()

        self.assertEqual(so.status, 40)
        self.assertEqual(so.status_text, 'Cancelled')


class ROTest(InvenTreeTestCase):
    """Unit tests for ReturnOrder models"""

    def test_ro_fields(self):
        """Check the OPTIONS endpoints for return order models"""

        field_names = order.ReturnOrder.fieldNames(self.api)

        names = [
            'contact',
            'line_items',
            'link',
            'overdue',
            'customer',
            'status',
            'reference'
        ]

        for name in names:
            self.assertIn(name, field_names)

    def test_ro_create(self):
        """Test that we can create a ReturnOrder"""

        customer = company.Company(self.api, pk=4)
        self.assertTrue(customer.is_customer)

        n = len(order.ReturnOrder.list(self.api))
        ref = f"RMA-00{n}"

        # Create a new ReturnOrder
        ro = customer.createReturnOrder(
            reference=ref,
            description="A new return order"
        )

        self.assertIsNotNone(ro)
        self.assertIsNotNone(ro.pk)

        # Check line items (should be zero)
        items = ro.getLineItems()
        self.assertEqual(len(items), 0)

        # Check extra line items (should be zero)
        items = ro.getExtraLineItems()
        self.assertEqual(len(items), 0)

        # Create some extra items
        for idx in range(3):
            ro.addExtraLineItem(
                reference=f"ref {idx}",
                notes="my notes",
            )

        self.assertEqual(len(ro.getExtraLineItems()), 3)

        for line in ro.getExtraLineItems():
            lo = line.getOrder()
            self.assertEqual(lo.pk, ro.pk)

        self.assertEqual(ro.getCustomer().pk, 4)
        self.assertIsNone(ro.getContact())

        # Test that we can "edit" the order
        self.assertEqual(ro.reference, ref)

        ro.save(data={
            'reference': 'RMA-99999',
        })

        ro.reload()
        self.assertEqual(ro.reference, 'RMA-99999')

        # Now, delete it
        ro.delete()

        # Should throw an error, as it no longer exists
        with self.assertRaises(HTTPError):
            ro.reload()

    def test_ro_cancel(self):
        """Test that an order can be cancelled"""

        ro = order.ReturnOrder.create(self.api, data={
            'description': 'To be cancelled',
            'customer': 4,
        })
        # Order should initially be 'pending'
        self.assertEqual(ro.status, 10)
        ro.cancel()
        ro.reload()
        # Order should now be 'cancelled'
        self.assertEqual(ro.status, 40)

    def test_ro_issue(self):
        """Test that an order can be issued"""

        ro = order.ReturnOrder.create(self.api, data={
            'description': 'To be issued',
            'customer': 4,
        })

        # Order should initially be 'pending'
        self.assertEqual(ro.status, 10)
        ro.issue()
        ro.reload()
        # Order should now be 'in progress'
        self.assertEqual(ro.status, 20)

    def test_ro_complete(self):
        """Test that an order can be completed"""

        ro = order.ReturnOrder.create(self.api, data={
            'description': 'To be completed',
            'customer': 4,
        })
        ro.issue()
        ro.complete()
        ro.reload()
        # Order should now be 'complete'
        self.assertEqual(ro.status, 30)



"""Unit tests for currency exchange support"""


import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.currency import CurrencyManager  # noqa: E402


class CurrencyTest(InvenTreeTestCase):
    """Tests for currency support"""

    def test_fetch_data(self):
        """Test that we can fetch currency data from the server"""

        mgr = CurrencyManager(self.api)
        mgr.updateFromServer()

        self.assertIsNotNone(mgr.base_currency)
        self.assertIsNotNone(mgr.exchange_rates)

    def test_base_currency(self):
        """Test call to 'getBaseCurrency'"""

        mgr = CurrencyManager(self.api)
        base = mgr.getBaseCurrency()

        self.assertEqual(base, 'USD')
        self.assertIsNotNone(mgr.exchange_rates)

    def test_exchange_rates(self):
        """Test call to 'getExchangeRates'"""

        mgr = CurrencyManager(self.api)
        rates = mgr.getExchangeRates()

        self.assertIsNotNone(rates)
        self.assertIsNotNone(mgr.base_currency)

    def test_conversion(self):
        """Test currency conversion"""

        mgr = CurrencyManager(self.api)
        mgr.updateFromServer()

        # Override data
        mgr.base_currency = 'USD'
        mgr.exchange_rates = {
            'USD': 1.00,
            'AUD': 1.40,
            'EUR': 0.90,
            'JPY': 128
        }

        test_sets = [
            (1, 'AUD', 'USD', 0.7143),
            (1, 'EUR', 'USD', 1.1111),
            (5, 'JPY', 'AUD', 0.0547),
            (5, 'AUD', 'EUR', 3.2143),
            (1, 'USD', 'JPY', 128)
        ]

        for value, source, target, result in test_sets:
            converted = mgr.convertCurrency(value, source, target)

            self.assertEqual(result, round(converted, 4))



# -*- coding: utf-8 -*-

import inventree.base


class ReportPrintingMixin:
    """Mixin class for report printing."""

    def getTemplateId(self, template):
        """Return the ID (pk) from the supplied template."""

        if type(template) in [str, int]:
            return int(template)
        
        if hasattr(template, 'pk'):
            return int(template.pk)
        
        raise ValueError(f"Provided report template is not a valid type: {type(template)}")

    def printReport(self, report, destination=None, *args, **kwargs):
        """Print the report belonging to the given item.

        Set the report with 'report' argument, as the ID of the corresponding
        report. A corresponding report object can also be given.

        The file will be downloaded to 'destination'.
        Use overwrite=True to overwrite an existing file.

        If neither plugin nor destination is given, nothing will be done
        """

        print_url = 'report/print/'
        template_id = self.getTemplateId(report)

        response = self._api.post(
            print_url,
            {
                'template': template_id,
                'items': [self.pk],
            }
        )

        output = response.get('output', None)

        if output and destination:
            return self._api.downloadFile(url=output, destination=destination, *args, **kwargs)
        else:
            return response

    def getReportTemplates(self, **kwargs):
        """Return a list of report templates which match this model class."""

        return ReportTemplate.list(self._api, model_type=self.getModelType(), **kwargs)


class ReportFunctions(inventree.base.MetadataMixin, inventree.base.InventreeObject):
    """Base class for report functions"""

    @classmethod
    def create(cls, api, data, template, **kwargs):
        """Create a new report by uploading a template file. Convenience wrapper around base create() method.

        Args:
            data: Dict of data including at least name and description for the template
            template: Either a string (filename) or a file object
        """

        try:
            # If template is already a readable object, don't convert it
            if template.readable() is False:
                raise ValueError("Template file must be readable")
        except AttributeError:
            template = open(template)
            if template.readable() is False:
                raise ValueError("Template file must be readable")

        try:
            response = super().create(api, data=data, files={'template': template}, **kwargs)
        finally:
            if template is not None:
                template.close()
        return response

    def save(self, data=None, template=None, **kwargs):
        """Save report data to database. Convenience wrapper around save() method.

        Args:
            data (optional): Dict of data to change for the template.
            template (optional): Either a string (filename) or a file object, to upload a new template
        """

        if template is not None:
            try:
                # If template is already a readable object, don't convert it
                if template.readable() is False:
                    raise ValueError("Template file must be readable")
            except AttributeError:
                template = open(template, 'r')
                if template.readable() is False:
                    raise ValueError("Template file must be readable")

            if 'files' in kwargs:
                files = kwargs.pop('kwargs')
                files['template'] = template
            else:
                files = {'template': template}
        else:
            files = None

        try:
            response = super().save(data=data, files=files)
        finally:
            if template is not None:
                template.close()
        return response

    def downloadTemplate(self, destination, overwrite=False):
        """Download template file for the report to the given destination"""

        # Use downloadFile method to get the file
        return self._api.downloadFile(url=self._data['template'], destination=destination, overwrite=overwrite)


class ReportTemplate(ReportFunctions):
    """Class representing the ReportTemplate model."""

    URL = 'report/template/'



"""
Unit test for basic model class functionality
"""

import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.base import InventreeObject  # noqa: E402


class BaseModelTests(InvenTreeTestCase):
    """Simple unit tests for the InvenTreeObject class"""

    def test_create(self):
        """Unit tests for InventreeObject creation"""

        # Test with non-int pk
        with self.assertRaises(TypeError):
            InventreeObject(None, pk='test')

        # Test with invalid pk
        for pk in [-1, 0]:
            with self.assertRaises(ValueError):
                InventreeObject(None, pk=pk)

        # Test with pk supplied in data
        with self.assertRaises(TypeError):
            InventreeObject(
                None,
                data={
                    'pk': 'seven',
                }
            )

    def test_data_access(self):
        """Test data access functionality"""

        obj = InventreeObject(
            None,
            data={
                "pk": "10",
                "hello": "world",
                "name": "My name",
                "description": "My description",
            }
        )

        # Test __getattr__ access
        self.assertEqual(obj.pk, 10)
        self.assertEqual(obj.name, "My name")

        with self.assertRaises(AttributeError):
            print(obj.doesNotExist)

        # Test __getitem__ access
        self.assertEqual(obj['description'], 'My description')
        self.assertEqual(obj['pk'], '10')
        self.assertEqual(obj['hello'], 'world')

        for k in ['fake', 'data', 'values']:
            with self.assertRaises(KeyError):
                print(obj[k])
